using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace XELF.Tool {

	public class Canvas : DrawableGameComponent {
		readonly Tools tools;
		readonly Pencil pencil;
		SpriteBatch spriteBatch;
		readonly HistoricalPalette historicalPalette;

		readonly TextureFromFile textureFromFile = new TextureFromFile();

		readonly List<Document> documents = new List<Document>();

		public HistoricalPalette HistoricalPalette { get { return historicalPalette; } }

		public Layer PickLayer(ref Point location) {
			Point screenLocation = location;
			for (int i = documents.Count; --i >= 0; ) {
				Document document = documents[i];
				Vector2 position = document.PointFromScreen(new Vector2(screenLocation.X, screenLocation.Y));
				location.X = (int)Math.Floor(position.X);
				location.Y = (int)Math.Floor(position.Y);
				Layer layer = document.PickLayer(ref location);
				if (layer != null) {
					return layer;
				}
			}
			location = screenLocation;
			return null;
		}

		public Layer HistoricalPaletteLayer {
			get { return documents[2].Layers[1]; }
		}
		public Layer PaletteLayer {
			get { return documents[2].CurrentLayer; }
		}
		public Layer CurrentLayer {
			get { return documents[1].CurrentLayer; }
		}
		public Layer BackgroundLayer {
			get { return documents[0].CurrentLayer; }
		}

		public Tools Tools { get { return tools; } }
		public Pencil Pencil { get { return pencil; } }

		public Canvas(Tools tools)
			: base(tools.Game) {
			this.tools = tools;

			Game.Components.Add(new PointerHUD(this));
			Game.Components.Add(pencil = new Pencil(this));
			Game.Components.Add(historicalPalette = new HistoricalPalette(this));

			documents.Add(new Document());
			documents.Add(new Document());
			documents.Add(new Document());
			documents[2].Scale = 16;
			documents[1].Scale = 8;
			documents[0].Scale = 8;

			documents[2].Translation = new Point(32, 32);
			documents[1].Translation = new Point(32, 64);
			documents[0].Translation = new Point(32, 64);

		}

		protected override void LoadContent() {
			spriteBatch = new SpriteBatch(GraphicsDevice);
			CreatePaletteLayer();
			CreateHistoricalPaletteLayer();
			Point size = new Point(64, 64);
			CreateLayer(size);
			documents[0].AddLayer(new Layer());
			SetDocumentSize(size);
			base.LoadContent();
		}

		void CreateHistoricalPaletteLayer() {
			Layer layer = new Layer();
			layer.Texture = new Texture2D(GraphicsDevice, 16, 1, 1, TextureUsage.None, SurfaceFormat.Vector4);
			layer.Location = new Point(10, 0);
			layer.Locked = true;
			documents[2].AddLayer(layer);
		}
		void CreatePaletteLayer() {
			Layer layer = new Layer();

			layer.SetImage(GraphicsDevice, new Vector4[]{
				new Vector4(0,0,0,0),
				new Vector4(0,0,0,1),
				new Vector4(0,0,1,1),
				new Vector4(0,1,0,1),
				new Vector4(0,1,1,1),
				new Vector4(1,0,0,1),
				new Vector4(1,0,1,1),
				new Vector4(1,1,0,1),
				new Vector4(1,1,1,1),
				}, new Point(9, 1));
			layer.Location = new Point(0, 0);
			layer.Locked = true;
			documents[2].AddLayer(layer);
		}

		Layer CreateLayer(Point size) {
			Layer layer = new Layer();
			layer.Texture = new Texture2D(GraphicsDevice,
				size.X, size.Y, 1, TextureUsage.None, SurfaceFormat.Vector4);
			documents[1].AddLayer(layer);
			return layer;
		}
		void SetDocumentSize(Point size) {
			documents[1].Size = size;
			documents[0].Size = size;

			Vector4[] image = new Vector4[size.X * size.Y];

			Vector4 c0 = new Vector4(0.25f, 0.25f, 0.25f, 1);
			Vector4 c1 = new Vector4(0.50f, 0.50f, 0.50f, 1);

			for (int y = 0; y < size.Y; y++) {
				for (int x = 0; x < size.X; x++) {
					image[x + y * size.X] = (((x + y) & 1) == 0) ? c0 : c1;
				}
			}

			BackgroundLayer.SetImage(GraphicsDevice, image, size);
		}

		void Load() {
			using (System.Windows.Forms.OpenFileDialog dialog = new System.Windows.Forms.OpenFileDialog()) {
				dialog.Filter
					= "Supported Files|*.ppm;*.pfm;*.tga;*.dds;*.dib;*.hdr;*.png;*.bmp;*.jpg;*.jpeg;*.emf;*.gif;*.exif;*.icon;*.tiff;*.wmf|"
					+ "Portable Network Graphics|*.png|"
					+ "Bitmap|*.bmp|"
					+ "Joint Photographic Experts Group|*.jpg;*.jpeg|"
					+ "Enhanced Meta File|*.emf|"
					+ "Graphics Interchange Format|*.gif|"
					+ "Exchangeable Image Format|*.exif|"
					+ "Windows Icon Image|*.icon|"
					+ "Tagged Image File Format|*.tiff|"
					+ "Windows Meta File|*.wmf|"
					+ "Portable Pixel Map|*.ppm|"
					+ "(*.pfm)|*.pfm|"
					+ "Truevision Advanced Raster Graphics Adapter|*.tga|"
					+ "(*.hdr)|*.hdr|"
					+ "DirectDraw Surface|*.dds|"
					+ "Device Independent Bitmap|*.dib";
				if (dialog.ShowDialog() == System.Windows.Forms.DialogResult.OK) {
					string extension = dialog.FileName.Substring(dialog.FileName.LastIndexOf('.') + 1);
					Texture2D texture = null;
					switch (extension) {
					default:
						texture = textureFromFile.Load(GraphicsDevice, dialog.FileName);
						break;
					case "ppm":
					case "pfm":
					case "tga":
					case "hdr":
					case "dds":
					case "dib": {
							TextureCreationParameters parameters =
								Texture2D.GetCreationParameters(GraphicsDevice, dialog.FileName);
							TextureInformation information = Texture.GetTextureInformation(dialog.FileName);
							parameters.Format = SurfaceFormat.Vector4;
							texture = Texture2D.FromFile(GraphicsDevice, dialog.FileName, parameters);
						}
						break;
					}
					SetDocumentSize(new Point(texture.Width, texture.Height));
					CurrentLayer.Texture = texture;
					documents[1].Path = dialog.FileName;
				}
			}
		}

		void Save() {
			using (System.Windows.Forms.SaveFileDialog dialog = new System.Windows.Forms.SaveFileDialog()) {
				dialog.FileName = documents[1].Path;
				dialog.Filter
					= "Portable Network Graphics|*.png|"
					+ "Joint Photographic Experts Group|*.jpg;*.jpeg|"
					+ "DirectDraw Surface|*.dds|"
					+ "Portable Pixel Map|*.ppm|"
					+ "(*.pfm)|*.pfm|"
					+ "Truevision Advanced Raster Graphics Adapter|*.tga|"
					+ "(*.hdr)|*.hdr|"
					+ "Bitmap|*.bmp|"
					+ "Device Independent Bitmap|*.dib";
				if (dialog.ShowDialog() == System.Windows.Forms.DialogResult.OK) {
					int index = dialog.FileName.LastIndexOf('.');
					string extension = dialog.FileName.Substring(index + 1);
					ImageFileFormat format;
					if (TextureFromFile.ImageFileFormatFromExtension(extension, out format)) {
						CurrentLayer.Texture.Save(dialog.FileName, format);
						documents[1].Path = dialog.FileName;
					}
				}
			}
		}

		public override void Update(GameTime gameTime) {
			if (tools.Keys.IsPress(Keys.OemPlus)) {
				documents[1].Scale *= 2;
				documents[0].Scale *= 2;
			}
			if (tools.Keys.IsPress(Keys.OemMinus)) {
				documents[1].Scale *= 0.5f;
				documents[0].Scale *= 0.5f;
			}
			if (tools.Keys.IsPress(Keys.Left)) {
				Mouse.SetPosition(tools.Pointer.MouseState.X - 1, tools.Pointer.MouseState.Y);
			}
			if (tools.Keys.IsPress(Keys.Right)) {
				Mouse.SetPosition(tools.Pointer.MouseState.X + 1, tools.Pointer.MouseState.Y);
			}
			if (tools.Keys.IsPress(Keys.Up)) {
				Mouse.SetPosition(tools.Pointer.MouseState.X, tools.Pointer.MouseState.Y - 1);
			}
			if (tools.Keys.IsPress(Keys.Down)) {
				Mouse.SetPosition(tools.Pointer.MouseState.X, tools.Pointer.MouseState.Y + 1);
			}
			if (tools.Keys.IsControlDown && tools.Keys.IsPress(Keys.O)) {
				Load();
			}
			if (tools.Keys.IsControlDown && tools.Keys.IsPress(Keys.S)) {
				Save();
			}
			base.Update(gameTime);
		}
		public override void Draw(GameTime gameTime) {
			spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Immediate, SaveStateMode.SaveState);
			GraphicsDevice.SamplerStates[0].MagFilter = TextureFilter.Point;

			foreach (Document document in documents) {
				int width = document.Size.X;
				int height = document.Size.Y;

				Vector2 center = new Vector2(width, height) * (document.Scale * 0.5f)
					+ new Vector2(document.Translation.X, document.Translation.Y);
				Vector2 origin = new Vector2(width, height) * 0.5f;
				foreach (Layer layer in document.Layers) {
					spriteBatch.Draw(layer.Texture,
						center + new Vector2(layer.Location.X, layer.Location.Y) * document.Scale,
						null, Color.White, 0, origin, document.Scale, SpriteEffects.None, 0);
				}
			}
			spriteBatch.End();
			base.Draw(gameTime);
		}
	}

}
