using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace XELF.Tool {

	public class Pointer : DrawableGameComponent {
		Vector2 position;
		MouseState previousState;
		MouseState state;
		public Pointer(Game game)
			: base(game) {
		}
		public MouseState PreviousMouseState { get { return previousState; } }
		public MouseState MouseState { get { return state; } }
		public override void Update(GameTime gameTime) {
			previousState = state;
			state = Mouse.GetState();
			position.X = state.X;
			position.Y = state.Y;
			base.Update(gameTime);
		}
		public Vector2 Position {
			get { return position; }
		}
	}

	public class Tools : DrawableGameComponent {
		SpriteFont font;
		readonly Pointer pointer;
		readonly KeyMap keys;

		public SpriteFont Font { get { return font; } }
		public Pointer Pointer { get { return pointer; } }
		public KeyMap Keys { get { return keys; } }


		public Tools(Game game)
			: base(game) {
			Game.Components.Add(pointer = new Pointer(game));
			Game.Components.Add(new MouseHUD(this));
			Game.Components.Add(keys = new KeyMap(game));
		}
		protected override void LoadContent() {
			font = Game.Content.Load<SpriteFont>("font/default");
			base.LoadContent();
		}
	}

	public class HistoricalPalette : DrawableGameComponent {
		readonly Canvas canvas;
		readonly Vector4[] image = new Vector4[16];

		public HistoricalPalette(Canvas canvas) : base(canvas.Game) { this.canvas = canvas; }

		protected override void LoadContent() {
			base.LoadContent();
		}

		public void Push(Vector4 color) {
			if (Array.Exists(image, delegate(Vector4 value) { return value == color; }))
				return;
			Array.Copy(image, 0, image, 1, image.Length - 1);
			image[0] = color;
			canvas.HistoricalPaletteLayer.SetImage(GraphicsDevice, image, new Point(image.Length, 1));
		}
	}

	public class PointerHUD : DrawableGameComponent {
		SpriteBatch spriteBatch;
		readonly Canvas canvas;
		readonly Vector4[] image = new Vector4[1];

		public PointerHUD(Canvas canvas)
			: base(canvas.Game) {
			this.canvas = canvas;
			DrawOrder = 0x10000000;
		}

		protected override void LoadContent() {
			spriteBatch = new SpriteBatch(GraphicsDevice);
			base.LoadContent();
		}
		public override void Draw(GameTime gameTime) {
			Vector2 position = canvas.Tools.Pointer.Position;
			Point location;
			location.X = (int)Math.Floor(position.X);
			location.Y = (int)Math.Floor(position.Y);
			Layer layer = canvas.PickLayer(ref location);

			if (layer != null) {
				GraphicsDevice.Textures[0] = null;
				layer.Texture.GetData(0, new Rectangle(location.X, location.Y, 1, 1), image, 0, 1);
				spriteBatch.Begin();
				spriteBatch.DrawString(canvas.Tools.Font,
					string.Format("Size=({0},{1})\n({2},{3}) : Color={4}",
					layer.Texture.Width, layer.Texture.Height,
					location.X, location.Y, image[0].ToString()),
					new Vector2(0x180, 0x200), Color.White);
				spriteBatch.End();
			}

			base.Draw(gameTime);
		}
	}

	public class Pencil : DrawableGameComponent {
		readonly Canvas canvas;
		readonly Vector4[] image = new Vector4[1 ] { new Vector4(1, 1, 1, 1) };

		public Pencil(Canvas canvas) : base(canvas.Game) { this.canvas = canvas; }

		public Vector4 Color { get { return image[0]; } }

		void Paint(Layer layer, Point location) {
			if (layer.Locked)
				return;
			layer.Texture.SetData(0, new Rectangle(location.X, location.Y, 1, 1),
				image, 0, 1, SetDataOptions.None);
		}
		void Pick(Layer layer, Point location) {
			layer.Texture.GetData(0, new Rectangle(location.X, location.Y, 1, 1),
				image, 0, 1);
			canvas.HistoricalPalette.Push(image[0]);
		}

		public override void Update(GameTime gameTime) {
			if (!Game.IsActive)
				return;
			if (canvas.Tools.Pointer.MouseState.LeftButton == ButtonState.Pressed
				|| canvas.Tools.Keys.IsDown(Keys.Z)) {
				Vector2 position = canvas.Tools.Pointer.Position;
				Point location;
				location.X = (int)Math.Floor(position.X);
				location.Y = (int)Math.Floor(position.Y);
				Layer layer = canvas.PickLayer(ref location);
				if (layer != null) Paint(layer, location);
			}
			if (canvas.Tools.Pointer.MouseState.RightButton == ButtonState.Pressed
				|| canvas.Tools.Keys.IsDown(Keys.X)) {
				Vector2 position = canvas.Tools.Pointer.Position;
				Point location;
				location.X = (int)Math.Floor(position.X);
				location.Y = (int)Math.Floor(position.Y);
				Layer layer = canvas.PickLayer(ref location);
				if (layer != null) Pick(layer, location);
			}
		}
	}

	public class KeyMap : GameComponent {
		KeyboardState previous;
		KeyboardState current;

		public KeyMap(Game game)
			: base(game) {
		}

		public override void Update(GameTime gameTime) {
			previous = current;
			current = Keyboard.GetState();
			base.Update(gameTime);
		}
		public bool IsPress(Keys key) {
			return current.IsKeyDown(key) && previous.IsKeyUp(key);
		}
		public bool IsDown(Keys key) {
			return current.IsKeyDown(key);
		}
		public bool IsControlDown {
			get {
				return current.IsKeyDown(Keys.LeftControl) || current.IsKeyDown(Keys.RightControl);
			}
		}
		public bool IsPress(params Keys[] keys) {
			foreach (Keys key in keys) {
				if (!IsPress(key))
					return false;
			}
			return true;
		}
	}

	public class MouseHUD : DrawableGameComponent {
		SpriteBatch spriteBatch;
		SpriteFont font;
		readonly Tools tools;

		public MouseHUD(Tools tools)
			: base(tools.Game) {
			this.tools = tools;
			DrawOrder = 0x10000000;
		}

		protected override void LoadContent() {
			spriteBatch = new SpriteBatch(GraphicsDevice);
			font = Game.Content.Load<SpriteFont>("font/default");
		}

		public override void Draw(GameTime gameTime) {
			MouseState mouseState = tools.Pointer.MouseState;
			bool left = mouseState.LeftButton == ButtonState.Pressed;
			bool right = mouseState.RightButton == ButtonState.Pressed;
			bool middle = mouseState.MiddleButton == ButtonState.Pressed;
			bool xbutton1 = mouseState.XButton1 == ButtonState.Pressed;
			bool xbutton2 = mouseState.XButton2 == ButtonState.Pressed;
			spriteBatch.Begin();
			spriteBatch.DrawString(font,
				string.Format("({0},{1} {2} {3} {4} {5} {6})", mouseState.X, mouseState.Y,
				left ? "L " : "- ",
				middle ? "M " : "- ",
				right ? "R " : "- ",
				xbutton1 ? "X1" : "- ",
				xbutton2 ? "X2" : "- "
				), Vector2.Zero, Color.White);
			spriteBatch.End();

			base.Draw(gameTime);
		}
	}

	public class ToolGame : Microsoft.Xna.Framework.Game {
		readonly GraphicsDeviceManager graphics;
		readonly Tools tools;
		readonly Canvas canvas;

		public ToolGame() {
			graphics = new GraphicsDeviceManager(this);
			graphics.SynchronizeWithVerticalRetrace = false;
			Content.RootDirectory = "Content";
			IsMouseVisible = true;
			TargetElapsedTime = TimeSpan.FromMilliseconds(1);
			IsFixedTimeStep = false;
			InactiveSleepTime = TimeSpan.FromMilliseconds(500);
			Window.AllowUserResizing = true;
			Components.Add(tools = new Tools(this));
			Components.Add(canvas = new Canvas(tools));
		}

		protected override void Draw(GameTime gameTime) {
			graphics.GraphicsDevice.Clear(Color.CornflowerBlue);
			base.Draw(gameTime);
		}
	}
}
